#!/usr/bin/perl -w
#
# WardenClientCommon.pm
#
# Copyright (C) 2011-2012 Cesnet z.s.p.o
#
# Use of this source is governed by a BSD-style license, see LICENSE file.

package WardenClientCommon;

use strict;
use Carp;

use Net::SSLeay qw(post_https make_headers make_form
                  load_error_strings
                  set_cert_and_key set_fd 
                  get_peer_certificate ssl_write_all
                  print_errs die_if_ssl_error die_now ssl_read_all);
use XML::Parser;

our $VERSION = "2.2";

#-------------------------------------------------------------------------------
# errMsg - prints error msg and returns undef or prints warning and returns 1
#-------------------------------------------------------------------------------
sub errMsg
{
  my $msg = shift;
  my $type = shift; 
  defined $type or $type = "err"; # default type is err. Other: warn
  
  # is this error report and is Verbose logging mode enabled?
  if (($type eq "err") && ($WardenClientConf::LOG_VERBOSE)) { # user wants to log debug information
    $msg .= "\nStack info: " . Carp::longmess();
  }

  # log into STDERR?
  if ($WardenClientConf::LOG_STDERR) {
    print STDERR $msg . "\n";
  }

  # log into Syslog?
  if ($WardenClientConf::LOG_SYSLOG) {
    openlog("Warden-client:", "pid", "$WardenClientConf::LOG_SYSLOG_FACILITY");
    syslog("$type|$WardenClientConf::LOG_SYSLOG_FACILITY", $msg . "\n");
    closelog();
  }

  if ($type eq 'warn') { # case of 'warn'
    return 1;
  } else { # case of 'err'
    return;
  }

} # End of errMsg

#-------------------------------------------------------------------------------
# https_send_receive - send $out_message, return (data, http error, headers)
#-------------------------------------------------------------------------------

sub https_send_receive
{
    my ($dest_serv, $port, $out_message, $crt_path, $key_path, $ca_path) = @_;
    my ($ctx, $ssl, $res, $errs);

    ($res, $errs) = Net::SSLeay::open_tcp_connection($dest_serv, $port);
    goto final unless $res;

    # SSLeay init
    load_error_strings();
    Net::SSLeay::SSLeay_add_ssl_algorithms();
    Net::SSLeay::randomize();

    # Setup and configure protocol context
    $ctx = Net::SSLeay::new_x_ctx();
    goto cleanup2 if $errs = print_errs('CTX_new') or !$ctx;
                                                                                                                                                                                                                                               
    Net::SSLeay::CTX_set_options($ctx, &Net::SSLeay::OP_ALL);
    goto cleanup2 if $errs = print_errs('CTX_set_options');

    # Setup certificates
    warn "Cert `$crt_path' given without key" if $crt_path && !$key_path;
    set_cert_and_key($ctx, $crt_path, $key_path) if $crt_path;

    # Set allowed CAs and require Warden server peer verify
    Net::SSLeay::CTX_load_verify_locations($ctx, $ca_path, '');
    goto cleanup2 if $errs = print_errs('CTX_load_verify_locations');
    Net::SSLeay::CTX_set_verify($ctx, &Net::SSLeay::VERIFY_PEER, 0);
    goto cleanup2 if $errs = print_errs('CTX_set_verify');
                                                                                                                                                                                                                                               
    # Create SSL session
    $ssl = Net::SSLeay::new($ctx);                                                                                                                                                                                                                          
    goto cleanup if $errs = print_errs('SSL_new') or !$ssl;

    # Attach filehandle and connect
    set_fd($ssl,  fileno(Net::SSLeay::SSLCAT_S));
    goto cleanup if $errs = print_errs('set_fd');

    $res = Net::SSLeay::connect($ssl);                                                                                                                                                                                                         
    goto cleanup_shut if $errs = print_errs('SSL_connect');

    # Ok, exchange data
    ($res, $errs) = ssl_write_all($ssl, $out_message);
    goto cleanup_shut unless $res;

    ($res, $errs) = ssl_read_all($ssl);

cleanup_shut:
    shutdown Net::SSLeay::SSLCAT_S, 2;

cleanup:
    Net::SSLeay::free ($ssl);
    $errs .= print_errs('SSL_free');

cleanup2:
    Net::SSLeay::CTX_free ($ctx);
    $errs .= print_errs('CTX_free');
    close Net::SSLeay::SSLCAT_S;

final:
    $res = "HTTP/1.0 900 NET OR SSL ERROR\r\n\r\n$errs\r\n\r\n$res" if $errs;

    my ($prolog, $data) = split /\s?\n\s?\n/, $res, 2;
    my ($ret, $headers) = split /\s?\n/, $prolog, 2;

    return wantarray ? ($data, $ret, $headers) : $data;
}

#-------------------------------------------------------------------------------
# expat2hash - convert convoluted expat list of lists to simpler hash
#-------------------------------------------------------------------------------
sub expat2hash {
  my $l = shift;

  # send XML attributes to the fiery pits of hell
  if (ref($l->[0]) eq "HASH") {
    shift(@{$l});
  }

  my $res = {};
  while (scalar(@{$l})!=0) {
    my $key = shift(@{$l});
    my $value = shift(@{$l});
    if (ref($value) eq "ARRAY") {
      # process sublists recursively
      $value = expat2hash($value);
    }
    push(@{$res->{$key}}, $value);
  };

  # flatten list singles 
  # special case warden events to simplify access, we want
  # them in list even if only single event arrives
  for my $key (keys %{$res}) {
    if (scalar(@{$res->{$key}})==1 and not ($key eq 'event')) {
      $res->{$key} = $res->{$key}->[0];
    }
  }

  # nil values yield empty hashes now
  if (!%{$res}) {
    undef $res;
  } 
  # flatten XML text node singles
  elsif (scalar(keys(%{$res}))==1 and exists $res->{'0'}) {
    $res = $res->{'0'};
  };

  return $res;
}

#-------------------------------------------------------------------------------
# hash2soapquery - convert parameters from hash into SOAP query payload
#-------------------------------------------------------------------------------
sub hash2soapquery {
  my ($server, $port, $service, $method, $data) = @_;
  my @datalist;

  while (my ($key, $value) = each %{$data} )
  {
    if (!defined $value) {
      push @datalist, "<$key xsi:nil=\"true\" />"
    } elsif ($value =~ /^[+-]?\d+\z/) {
      push @datalist, "<$key xsi:type=\"xsd:int\">$value</$key>"
    } else {
      push @datalist, "<$key xsi:type=\"xsd:string\">$value</$key>"
    }
  }

  my $payload = 
    '<?xml version="1.0" encoding="UTF-8"?>' .
    '<soap:Envelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:soapenc="http://schemas.xmlsoap.org/soap/encoding/" xmlns:xsd="http://www.w3.org/2001/XMLSchema" soap:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/" xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/">' .
      '<soap:Body>' .
        "<$method xmlns=\"Warden\">" .
          '<request>' .
          join("", @datalist) .
          '</request>' .
        "</$method>" .
      '</soap:Body>' .
    '</soap:Envelope>';

  my $paylen = length($payload);
  
  my $header = 
    "POST /$service HTTP/1.1\r\n" .
    "Host: $server:$port\r\n" .
    "Content-Type: application/soap+xml; charset=utf-8\r\n" .
    "Content-Length: $paylen\r\n" .
    "SOAPAction: \"http://www.w3.org/2003/05/soap-envelope\"\r\n" .
    "\r\n";

  return $header . $payload;
}


#-------------------------------------------------------------------------------
# c2s - connect to server, send request and receive response
#-------------------------------------------------------------------------------
sub c2s
{
  my $uri               = shift;
  my $ssl_key_file      = shift;
  my $ssl_cert_file     = shift;
  my $ssl_ca_file       = shift;
  my $method            = shift;
  my $data              = shift;

  my $client;
  my ($server, $port, $service) = $uri =~ /https:\/\/(.+)\:(\d+)\/(.+)/;

  # format HTTP request
  my $message = hash2soapquery($server, $port, $service, $method, $data);

  my ($res, $err);

  eval {
    ($res, $err, undef) = https_send_receive($server, $port, $message, $ssl_cert_file, $ssl_key_file, $ssl_ca_file)
  } or return errMsg('Unknown error in c2s() sending SOAP data, ' . $@);

  # check HTTP error
  my ($code) = $err =~ /[^ ]* ([^ ]*) /;
  if ($code != "200") {
    return errMsg("Server returned HTTP error: $err.");
  }
  
  # check server response
  if (!defined $res) {
    return errMsg("Server returned empty response. Problem with used SSL ceritificates or Warden server at $server:$port is down.");
  }
  # deserialize response from server
  my $result;
  eval {
    my $parser = XML::Parser->new(Style => "Tree", ErrorContext => 2);
    $result = expat2hash($parser->parse($res));
  } or return errMsg('Unknown error in SOAP data deserialization. Received data: ' . $res . ', ' . $@);
  
  # Flatten unneeded branches if structure is sane
  unless (exists($result->{'soap:Envelope'}) and exists($result->{'soap:Envelope'}->{'soap:Body'})) {
     return errMsg('Server returned unexpected SOAP response. Received data: ' . $res);
  }
  $result = $result->{'soap:Envelope'}->{'soap:Body'};

  return $result;
}

#-------------------------------------------------------------------------------
# getClientsInfo - retrieve information about other clients from Warden server
#-------------------------------------------------------------------------------
sub getClientsInfo 
{
  my $warden_path = shift;

  my $etcdir = $warden_path . "/etc/";
  my $libdir = $warden_path . "/lib/";

  require $libdir . "WardenClientConf.pm";

  # read the config file
  my $conf_file = $etcdir . "warden-client.conf";
  WardenClientConf::loadConf($conf_file);

  # c2s() returns undef on fail
  my $response = c2s($WardenClientConf::URI, $WardenClientConf::SSL_KEY_FILE, $WardenClientConf::SSL_CERT_FILE, $WardenClientConf::SSL_CA_FILE, "getClientInfo");
  
  defined $response or return; # receive data or return undef     

  # parse returned SOAP data object with clients
  my @clients;
  my ($client_id, $hostname, $registered, $requestor, $service, $client_type, $type, $receive_own_events, $description_tags, $ip_net_client);
  my @response_list = $response->valueof('/Envelope/Body/getClientInfoResponse/client/');

  while (scalar @response_list) {
    my $response_data = shift(@response_list);
    my %client;

    $client{'client_id'}		= $response_data->{'CLIENT_ID'} ;
    $client{'hostname'}			= $response_data->{'HOSTNAME'};
    $client{'registered'}		= $response_data->{'REGISTERED'};
    $client{'requestor'}		= $response_data->{'REQUESTOR'};
    $client{'service'}	 		= $response_data->{'SERVICE'};
    $client{'client_type'}		= $response_data->{'CLIENT_TYPE'};
    $client{'type'}			= $response_data->{'TYPE'};
    $client{'receive_own_events'}	= $response_data->{'RECEIVE_OWN_EVENTS'};
    $client{'description_tags'}		= $response_data->{'DESCRIPTION_TAGS'};
    $client{'ip_net_client'}		= $response_data->{'IP_NET_CLIENT'};

    # push received clients from warden server into @clients which is returned
    push (@clients,\%client);
  }

  return @clients;
}

1;
